"""Unit test helper methods.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import os

import pytest
from werkzeug.http import dump_cookie

from incorporealcms import create_app
from incorporealcms.pages import (generate_parent_navs, instance_resource_path_to_request_path, render,
                                  request_path_to_breadcrumb_display, request_path_to_instance_resource_path)

HERE = os.path.dirname(os.path.abspath(__file__))


def test_generate_page_navs_index(app):
    """Test that the index page has navs to the root (itself)."""
    with app.app_context():
        assert generate_parent_navs('pages/index.md') == [('example.com', '/')]


def test_generate_page_navs_subdir_index(app):
    """Test that dir pages have navs to the root and themselves."""
    with app.app_context():
        assert generate_parent_navs('pages/subdir/index.md') == [('example.com', '/'), ('subdir', '/subdir/')]


def test_generate_page_navs_subdir_real_page(app):
    """Test that real pages have navs to the root, their parent, and themselves."""
    with app.app_context():
        assert generate_parent_navs('pages/subdir/page.md') == [('example.com', '/'), ('subdir', '/subdir/'),
                                                                ('Page', '/subdir/page')]


def test_generate_page_navs_subdir_with_title_parsing_real_page(app):
    """Test that title metadata is used in the nav text."""
    with app.app_context():
        assert generate_parent_navs('pages/subdir-with-title/page.md') == [
            ('example.com', '/'),
            ('SUB!', '/subdir-with-title/'),
            ('page', '/subdir-with-title/page')
        ]


def test_generate_page_navs_subdir_with_no_index(app):
    """Test that breadcrumbs still generate even if a subdir doesn't have an index.md."""
    with app.app_context():
        assert generate_parent_navs('pages/no-index-dir/page.md') == [
            ('example.com', '/'),
            ('/no-index-dir/', '/no-index-dir/'),
            ('page', '/no-index-dir/page')
        ]


def test_render_with_user_dark_theme(app):
    """Test that a request with the dark theme selected renders the dark theme."""
    cookie = dump_cookie("user-style", 'dark')
    with app.test_request_context(headers={'COOKIE': cookie}):
        assert b'/static/css/dark.css' in render('base.html').data
        assert b'/static/css/light.css' not in render('base.html').data


def test_render_with_user_light_theme(app):
    """Test that a request with the light theme selected renders the light theme."""
    with app.test_request_context():
        assert b'/static/css/light.css' in render('base.html').data
        assert b'/static/css/dark.css' not in render('base.html').data


def test_render_with_no_user_theme(app):
    """Test that a request with no theme set renders the light theme."""
    with app.test_request_context():
        assert b'/static/css/light.css' in render('base.html').data
        assert b'/static/css/dark.css' not in render('base.html').data


def test_render_with_theme_defaults_affects_html(app):
    """Test that the base themes are all that's presented in the HTML."""
    # test we can remove stuff from the default
    with app.test_request_context():
        assert b'?style=light' in render('base.html').data
        assert b'?style=dark' in render('base.html').data
        assert b'?style=plain' in render('base.html').data


def test_render_with_theme_overrides_affects_html(app):
    """Test that the overridden themes are presented in the HTML."""
    # test we can remove stuff from the default
    restyled_app = create_app(instance_path=os.path.join(HERE, 'instance'),
                              test_config={'PAGE_STYLES': {'light': '/static/css/light.css'}})
    with restyled_app.test_request_context():
        assert b'?style=light' in render('base.html').data
        assert b'?style=dark' not in render('base.html').data
        assert b'?style=plain' not in render('base.html').data

    # test that we can add new stuff too/instead
    restyled_app = create_app(instance_path=os.path.join(HERE, 'instance'),
                              test_config={'PAGE_STYLES': {'cool': '/static/css/cool.css',
                                                           'warm': '/static/css/warm.css'},
                                           'DEFAULT_PAGE_STYLE': 'warm'})
    with restyled_app.test_request_context():
        assert b'?style=cool' in render('base.html').data
        assert b'?style=warm' in render('base.html').data


def test_render_with_theme_overrides(app):
    """Test that the loaded themes can be overridden from the default."""
    cookie = dump_cookie("user-style", 'cool')
    restyled_app = create_app(instance_path=os.path.join(HERE, 'instance'),
                              test_config={'PAGE_STYLES': {'cool': '/static/css/cool.css',
                                                           'warm': '/static/css/warm.css'}})
    with restyled_app.test_request_context(headers={'COOKIE': cookie}):
        assert b'/static/css/cool.css' in render('base.html').data
        assert b'/static/css/warm.css' not in render('base.html').data


def test_render_with_theme_overrides_not_found_is_default(app):
    """Test that theme overrides work, and if a requested theme doesn't exist, the default is loaded."""
    cookie = dump_cookie("user-style", 'nonexistent')
    restyled_app = create_app(instance_path=os.path.join(HERE, 'instance'),
                              test_config={'PAGE_STYLES': {'cool': '/static/css/cool.css',
                                                           'warm': '/static/css/warm.css'},
                                           'DEFAULT_PAGE_STYLE': 'warm'})
    with restyled_app.test_request_context(headers={'COOKIE': cookie}):
        assert b'/static/css/warm.css' in render('base.html').data
        assert b'/static/css/nonexistent.css' not in render('base.html').data


def test_request_path_to_instance_resource_path(app):
    """Test a normal URL request is transformed into the file path."""
    with app.test_request_context():
        assert request_path_to_instance_resource_path('index') == ('pages/index.md', 'markdown')


def test_request_path_to_instance_resource_path_direct_file(app):
    """Test a normal URL request is transformed into the file path."""
    with app.test_request_context():
        assert request_path_to_instance_resource_path('no-title') == ('pages/no-title.md', 'markdown')


def test_request_path_to_instance_resource_path_in_subdir(app):
    """Test a normal URL request is transformed into the file path."""
    with app.test_request_context():
        assert request_path_to_instance_resource_path('subdir/page') == ('pages/subdir/page.md', 'markdown')


def test_request_path_to_instance_resource_path_subdir_index(app):
    """Test a normal URL request is transformed into the file path."""
    with app.test_request_context():
        assert request_path_to_instance_resource_path('subdir/') == ('pages/subdir/index.md', 'markdown')


def test_request_path_to_instance_resource_path_relatives_walked(app):
    """Test a normal URL request is transformed into the file path."""
    with app.test_request_context():
        assert (request_path_to_instance_resource_path('subdir/more-subdir/../../more-metadata') ==
                ('pages/more-metadata.md', 'markdown'))


def test_request_path_to_instance_resource_path_relatives_walked_indexes_work_too(app):
    """Test a normal URL request is transformed into the file path."""
    with app.test_request_context():
        assert request_path_to_instance_resource_path('subdir/more-subdir/../../') == ('pages/index.md', 'markdown')


def test_request_path_to_instance_resource_path_relatives_walked_into_subdirs_also_fine(app):
    """Test a normal URL request is transformed into the file path."""
    with app.test_request_context():
        assert (request_path_to_instance_resource_path('subdir/more-subdir/../../subdir/page') ==
                ('pages/subdir/page.md', 'markdown'))


def test_request_path_to_instance_resource_path_permission_error_on_ref_above_pages(app):
    """Test that attempts to get above the base dir ("/../../foo") fail."""
    with app.test_request_context():
        with pytest.raises(PermissionError):
            assert request_path_to_instance_resource_path('../unreachable')


def test_request_path_to_instance_resource_path_isadirectory_on_file_like_req_for_dir(app):
    """Test that a request for e.g. '/foo' when foo is a dir indicate to redirect."""
    with app.test_request_context():
        with pytest.raises(IsADirectoryError):
            assert request_path_to_instance_resource_path('subdir')


def test_request_path_to_instance_resource_path_actual_file(app):
    """Test that a request for e.g. '/foo.png' when foo.png is a real file works."""
    with app.test_request_context():
        assert (request_path_to_instance_resource_path('bss-square-no-bg.png') ==
                ('pages/bss-square-no-bg.png', 'file'))


def test_request_path_to_instance_resource_path_markdown_symlink(app):
    """Test that a request for e.g. '/foo' when foo.md is a symlink to another .md file redirects."""
    with app.test_request_context():
        assert (request_path_to_instance_resource_path('symlink-to-no-title') ==
                ('pages/no-title.md', 'symlink'))


def test_request_path_to_instance_resource_path_file_symlink(app):
    """Test that a request for e.g. '/foo' when foo.txt is a symlink to another .txt file redirects."""
    with app.test_request_context():
        assert (request_path_to_instance_resource_path('symlink-to-foo.txt') ==
                ('pages/foo.txt', 'symlink'))


def test_request_path_to_instance_resource_path_dir_symlink(app):
    """Test that a request for e.g. '/foo' when /foo is a symlink to /bar redirects."""
    with app.test_request_context():
        assert (request_path_to_instance_resource_path('symlink-to-subdir/') ==
                ('pages/subdir', 'symlink'))


def test_request_path_to_instance_resource_path_subdir_symlink(app):
    """Test that a request for e.g. '/foo/baz' when /foo is a symlink to /bar redirects."""
    with app.test_request_context():
        assert (request_path_to_instance_resource_path('symlink-to-subdir/page-no-title') ==
                ('pages/subdir/page-no-title.md', 'symlink'))


def test_request_path_to_instance_resource_path_nonexistant_file_errors(app):
    """Test that a request for something not on disk errors."""
    with app.test_request_context():
        with pytest.raises(FileNotFoundError):
            assert request_path_to_instance_resource_path('nthanpthpnh')


def test_request_path_to_instance_resource_path_absolute_file_errors(app):
    """Test that a request for something not on disk errors."""
    with app.test_request_context():
        with pytest.raises(PermissionError):
            assert request_path_to_instance_resource_path('/etc/hosts')


def test_instance_resource_path_to_request_path_on_index(app):
    """Test index.md -> /."""
    with app.test_request_context():
        assert instance_resource_path_to_request_path('index.md') == ''


def test_instance_resource_path_to_request_path_on_page(app):
    """Test no-title.md -> no-title."""
    with app.test_request_context():
        assert instance_resource_path_to_request_path('no-title.md') == 'no-title'


def test_instance_resource_path_to_request_path_on_subdir(app):
    """Test subdir/index.md -> subdir/."""
    with app.test_request_context():
        assert instance_resource_path_to_request_path('subdir/index.md') == 'subdir/'


def test_instance_resource_path_to_request_path_on_subdir_and_page(app):
    """Test subdir/page.md -> subdir/page."""
    with app.test_request_context():
        assert instance_resource_path_to_request_path('subdir/page.md') == 'subdir/page'


def test_request_resource_request_root(app):
    """Test that a request can resolve to a resource and back to a request."""
    with app.test_request_context():
        instance_path, _ = request_path_to_instance_resource_path('index')
        instance_resource_path_to_request_path(instance_path) == ''


def test_request_resource_request_page(app):
    """Test that a request can resolve to a resource and back to a request."""
    with app.test_request_context():
        instance_path, _ = request_path_to_instance_resource_path('no-title')
        instance_resource_path_to_request_path(instance_path) == 'no-title'


def test_request_path_to_breadcrumb_display_patterns():
    """Test various conversions from request path to leaf nodes for display in the breadcrumbs."""
    assert request_path_to_breadcrumb_display('/foo') == 'foo'
    assert request_path_to_breadcrumb_display('/foo/') == 'foo'
    assert request_path_to_breadcrumb_display('/foo/bar') == 'bar'
    assert request_path_to_breadcrumb_display('/foo/bar/') == 'bar'
    assert request_path_to_breadcrumb_display('/') == ''



"""Test page requests.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import re
from unittest.mock import patch


def test_page_that_exists(client):
    """Test that the app can serve a basic file at the index."""
    response = client.get('/')
    assert response.status_code == 200
    assert b'<h1 id="test-index">test index</h1>' in response.data


def test_direct_file_that_exists(client):
    """Test that the app can serve a basic file at the index."""
    response = client.get('/foo.txt')
    assert response.status_code == 200
    assert b'test file' in response.data


def test_page_that_doesnt_exist(client):
    """Test that the app returns 404 for nonsense requests and they use my error page."""
    response = client.get('/ohuesthaoeusth')
    assert response.status_code == 404
    assert b'<b><tt>/ohuesthaoeusth</tt></b> does not seem to exist' in response.data
    # test the contact email config
    assert b'admin@example.com' in response.data


def test_files_outside_pages_do_not_get_served(client):
    """Test that page pathing doesn't break out of the instance/pages/ dir, and the error uses my error page."""
    response = client.get('/../unreachable')
    assert response.status_code == 400
    assert b'You\'re doing something you\'re not supposed to. Stop it?' in response.data


def test_internal_server_error_serves_error_page(client):
    """Test that various exceptions serve up the 500 page."""
    response = client.get('/actually-a-png')
    assert response.status_code == 500
    assert b'INTERNAL SERVER ERROR' in response.data
    # test the contact email config
    assert b'admin@example.com' in response.data


def test_oserror_is_500(client, app):
    """Test that an OSError raises as a 500."""
    with app.test_request_context():
        with patch('flask.current_app.open_instance_resource', side_effect=OSError):
            response = client.get('/')
            assert response.status_code == 500
            assert b'INTERNAL SERVER ERROR' in response.data


def test_unsupported_file_type_is_500(client, app):
    """Test a coding condition mishap raises as a 500."""
    with app.test_request_context():
        with patch('incorporealcms.pages.request_path_to_instance_resource_path', return_value=('foo', 'bar')):
            response = client.get('/')
            assert response.status_code == 500
            assert b'INTERNAL SERVER ERROR' in response.data


def test_weird_paths_do_not_get_served(client):
    """Test that we clean up requests as desired."""
    response = client.get('/../../')
    assert response.status_code == 400


def test_page_with_title_metadata(client):
    """Test that a page with title metadata has its title written."""
    response = client.get('/')
    assert response.status_code == 200
    assert b'<title>Index - example.com</title>' in response.data


def test_page_without_title_metadata(client):
    """Test that a page without title metadata gets the default title."""
    response = client.get('/no-title')
    assert response.status_code == 200
    assert b'<title>/no-title - example.com</title>' in response.data


def test_page_in_subdir_without_title_metadata(client):
    """Test that the title-less page display is as expected."""
    response = client.get('/subdir//page-no-title')
    assert response.status_code == 200
    assert b'<title>/subdir/page-no-title - example.com</title>' in response.data


def test_page_with_card_metadata(client):
    """Test that a page with opengraph metadata."""
    response = client.get('/more-metadata')
    assert response.status_code == 200
    assert b'<meta property="og:title" content="title for the page - example.com">' in response.data
    assert b'<meta property="og:description" content="description of this page made even longer">' in response.data
    assert b'<meta property="og:image" content="http://buh.com/test.img">' in response.data


def test_page_with_card_title_even_when_no_metadata(client):
    """Test that a page without metadata still has a card with the derived title."""
    response = client.get('/no-title')
    assert response.status_code == 200
    assert b'<meta property="og:title" content="/no-title - example.com">' in response.data
    assert b'<meta property="og:description"' not in response.data
    assert b'<meta property="og:image"' not in response.data


def test_page_with_forced_empty_title_just_shows_suffix(client):
    """Test that if a page specifies a blank Title meta tag explicitly, only the suffix is used in the title."""
    response = client.get('/forced-no-title')
    assert response.status_code == 200
    assert b'<title>example.com</title>' in response.data


def test_page_with_redirect_meta_url_redirects(client):
    """Test that if a page specifies a URL to redirect to, that the site serves up a 301."""
    response = client.get('/redirect')
    assert response.status_code == 301
    assert response.location == 'http://www.google.com/'


def test_page_has_modified_timestamp(client):
    """Test that pages have modified timestamps in them."""
    response = client.get('/')
    assert response.status_code == 200
    assert re.search(r'Last modified: ....-..-.. ..:..:.. ...', response.data.decode()) is not None


def test_that_page_request_redirects_to_directory(client):
    """Test that a request to /foo reirects to /foo/, if foo is a directory.

    This might be useful in cases where a formerly page-only page has been
    converted to a directory with subpages.
    """
    response = client.get('/subdir')
    assert response.status_code == 301
    assert response.location == '/subdir/'


def test_that_request_to_symlink_redirects_markdown(client):
    """Test that a request to /foo redirects to /what-foo-points-at."""
    response = client.get('/symlink-to-no-title')
    assert response.status_code == 301
    assert response.location == '/no-title'


def test_that_request_to_symlink_redirects_file(client):
    """Test that a request to /foo.txt redirects to /what-foo-points-at.txt."""
    response = client.get('/symlink-to-foo.txt')
    assert response.status_code == 301
    assert response.location == '/foo.txt'


def test_that_request_to_symlink_redirects_directory(client):
    """Test that a request to /foo/ redirects to /what-foo-points-at/."""
    response = client.get('/symlink-to-subdir/')
    assert response.status_code == 301
    assert response.location == '/subdir'
    # sadly, this location also redirects
    response = client.get('/subdir')
    assert response.status_code == 301
    assert response.location == '/subdir/'
    # but we do get there
    response = client.get('/subdir/')
    assert response.status_code == 200


def test_that_request_to_symlink_redirects_subdirectory(client):
    """Test that a request to /foo/bar redirects to /what-foo-points-at/bar."""
    response = client.get('/symlink-to-subdir/page-no-title')
    assert response.status_code == 301
    assert response.location == '/subdir/page-no-title'
    response = client.get('/subdir/page-no-title')
    assert response.status_code == 200


def test_that_dir_request_does_not_redirect(client):
    """Test that a request to /foo/ serves the index page, if foo is a directory."""
    response = client.get('/subdir/')
    assert response.status_code == 200
    assert b'another page' in response.data


def test_setting_selected_style_includes_cookie(client):
    """Test that a request with style=foo sets the cookie and renders appropriately."""
    response = client.get('/')
    style_cookie = client.get_cookie('user-style')
    assert style_cookie is None

    response = client.get('/?style=light')
    style_cookie = client.get_cookie('user-style')
    assert response.status_code == 200
    assert b'/static/css/light.css' in response.data
    assert b'/static/css/dark.css' not in response.data
    assert style_cookie.value == 'light'

    response = client.get('/?style=dark')
    style_cookie = client.get_cookie('user-style')
    assert response.status_code == 200
    assert b'/static/css/dark.css' in response.data
    assert b'/static/css/light.css' not in response.data
    assert style_cookie.value == 'dark'


def test_pages_can_supply_alternate_templates(client):
    """Test that pages can supply templates other than the default."""
    response = client.get('/')
    assert b'class="site-wrap site-wrap-normal-width"' in response.data
    assert b'class="site-wrap site-wrap-double-width"' not in response.data
    response = client.get('/custom-template')
    assert b'class="site-wrap site-wrap-normal-width"' not in response.data
    assert b'class="site-wrap site-wrap-double-width"' in response.data


def test_extra_footer_per_page(client):
    """Test that we don't include the extra-footer if there isn't one (or do if there is)."""
    response = client.get('/')
    assert b'<div class="extra-footer">' not in response.data
    response = client.get('/index-but-with-footer')
    assert b'<div class="extra-footer"><i>ooo <a href="a">a</a></i>' in response.data


def test_serving_static_files(client):
    """Test the usage of send_from_directory to serve extra static files."""
    response = client.get('/custom-static/css/warm.css')
    assert response.status_code == 200

    # can't serve directories, just files
    response = client.get('/custom-static/')
    assert response.status_code == 404
    response = client.get('/custom-static/css/')
    assert response.status_code == 404
    response = client.get('/custom-static/css')
    assert response.status_code == 404

    # can't serve files that don't exist or bad paths
    response = client.get('/custom-static/css/cold.css')
    assert response.status_code == 404
    response = client.get('/custom-static/css/../../unreachable.md')
    assert response.status_code == 404



"""Create the test app and other fixtures.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import os

import pytest

from incorporealcms import create_app

HERE = os.path.dirname(os.path.abspath(__file__))


@pytest.fixture
def app():
    """Create the Flask application, with test settings."""
    app = create_app(instance_path=os.path.join(HERE, 'instance'))

    yield app


@pytest.fixture
def client(app):
    """Create a test client based on the test app."""
    return app.test_client()



"""An application for running my Markdown-based sites.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import logging
import os
from logging.config import dictConfig

from flask import Flask, request


def create_app(instance_path=None, test_config=None):
    """Create the Flask app, with allowances for customizing path and test settings."""
    app = Flask(__name__, instance_relative_config=True, instance_path=instance_path)

    # if it doesn't already exist, create the instance folder
    os.makedirs(app.instance_path, exist_ok=True)

    # load defaults from config provided with the application
    app.config.from_object('incorporealcms.config.Config')
    # load specific instance configurations
    app.config.from_pyfile('config.py', silent=True)
    if test_config:
        app.config.from_mapping(test_config)

    dictConfig(app.config['LOGGING'])

    logger = logging.getLogger(__name__)

    logger.debug("instance path: %s", app.instance_path)

    @app.before_request
    def log_request():
        logger.info("REQUEST:  %s %s", request.method, request.path)

    @app.after_request
    def log_response(response):
        logger.info("RESPONSE: %s %s: %s", request.method, request.path, response.status)
        return response

    from . import error_pages, feed, pages, static
    app.register_blueprint(feed.bp)
    app.register_blueprint(pages.bp)
    app.register_blueprint(static.bp)
    app.register_error_handler(400, error_pages.bad_request)
    app.register_error_handler(404, error_pages.page_not_found)
    app.register_error_handler(500, error_pages.internal_server_error)

    return app



"""Serve dot diagrams inline.

SPDX-FileCopyrightText: © 2021 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import base64
import logging
import re

import markdown
import pydot

logger = logging.getLogger(__name__)


class InlinePydot(markdown.Extension):
    """Wrap the markdown prepcoressor."""

    def extendMarkdown(self, md):
        """Add InlinePydotPreprocessor to the Markdown instance."""
        md.preprocessors.register(InlinePydotPreprocessor(md), 'dot_block', 100)


class InlinePydotPreprocessor(markdown.preprocessors.Preprocessor):
    """Identify dot codeblocks and run them through pydot."""

    BLOCK_RE = re.compile(r'~~~{\s+pydot:(?P<filename>[^\s]+)\s+}\n(?P<content>.*?)~~~', re.DOTALL)

    def run(self, lines):
        """Match and generate diagrams from dot code blocks."""
        text = '\n'.join(lines)
        out = text
        for block_match in self.BLOCK_RE.finditer(text):
            filename = block_match.group(1)
            dot_string = block_match.group(2)
            logger.debug("matched markdown block: %s", dot_string)
            logger.debug("match start/end: %s/%s", block_match.start(), block_match.end())

            # use pydot to turn the text into pydot
            graphs = pydot.graph_from_dot_data(dot_string)
            if not graphs:
                logger.debug("some kind of issue with parsed 'dot' %s", dot_string)
                raise ValueError("error parsing dot text!")

            # encode the image and provide as an inline image in markdown
            encoded_image = base64.b64encode(graphs[0].create_png()).decode('ascii')
            data_path = f'data:image/png;base64,{encoded_image}'
            inline_image = f'![{filename}]({data_path})'

            # replace the image in the output markdown
            out = out.replace(block_match.group(0), inline_image)

        return out.split('\n')


def makeExtension(*args, **kwargs):
    """Provide the extension to the markdown extension loader."""
    return InlinePydot(*args, **kwargs)



"""Miscellaneous helper functions and whatnot.

SPDX-FileCopyrightText: © 2021 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import datetime
import logging
import os
import re

import markdown
from flask import current_app as app
from flask import make_response, render_template, request
from markupsafe import Markup

logger = logging.getLogger(__name__)


def get_meta_str(md, key):
    """Provide the page's (parsed in Markup obj md) metadata for the specified key, or '' if unset."""
    return " ".join(md.Meta.get(key)) if md.Meta.get(key) else ""


def init_md():
    """Initialize the Markdown parser.

    This used to done at the app level in __init__, but extensions like footnotes apparently
    assume the parser to only live for the length of parsing one document, and create double
    footnote ref links if the one parser sees the same document multiple times.
    """
    # initialize markdown parser from config, but include
    # extensions our app depends on, like the meta extension
    return markdown.Markdown(extensions=app.config['MARKDOWN_EXTENSIONS'] + ['meta'],
                             extension_configs=app.config['MARKDOWN_EXTENSION_CONFIGS'])


def instance_resource_path_to_request_path(path):
    """Reverse a (presumed to exist) RELATIVE disk path to the canonical path that would show up in a Flask route.

    This does not include the leading /, so aside from the root index case, this should be
    bidirectional.
    """
    return re.sub(r'^pages/', '', re.sub(r'.md$', '', re.sub(r'index.md$', '', path)))


def parse_md(resolved_path):
    """Given a file to parse, return file content and other derived data along with the md object."""
    try:
        logger.debug("opening resolved path '%s'", resolved_path)
        with app.open_instance_resource(resolved_path, 'r') as entry_file:
            mtime = datetime.datetime.fromtimestamp(os.path.getmtime(entry_file.name), tz=datetime.timezone.utc)
            entry = entry_file.read()
        logger.debug("resolved path '%s' read", resolved_path)
        md = init_md()
        content = Markup(md.convert(entry))
    except OSError:
        logger.exception("resolved path '%s' could not be opened!", resolved_path)
        raise
    except ValueError:
        logger.exception("error parsing/rendering markdown!")
        raise
    except TypeError:
        logger.exception("error loading/rendering markdown!")
        raise

    logger.debug("file metadata: %s", md.Meta)

    page_name = (get_meta_str(md, 'title') if md.Meta.get('title') else
                 f'/{instance_resource_path_to_request_path(resolved_path)}')
    page_title = f'{page_name} - {app.config["TITLE_SUFFIX"]}' if page_name else app.config['TITLE_SUFFIX']
    logger.debug("title (potentially derived): %s", page_title)

    return content, md, page_name, page_title, mtime


def render(template_name_or_list, **context):
    """Wrap Flask's render_template.

    * Determine the proper site theme to use in the template and provide it.
    """
    page_styles = app.config['PAGE_STYLES']
    selected_style = request.args.get('style', None)
    if selected_style:
        user_style = selected_style
    else:
        user_style = request.cookies.get('user-style')
        logger.debug("user style cookie: %s", user_style)
    context['user_style'] = page_styles.get(user_style, page_styles.get(app.config['DEFAULT_PAGE_STYLE']))
    context['page_styles'] = page_styles

    resp = make_response(render_template(template_name_or_list, **context))
    if selected_style:
        resp.set_cookie('user-style', selected_style)
    return resp



"""General page functionality.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import logging
import os

from flask import Blueprint, abort
from flask import current_app as app
from flask import redirect, request, send_from_directory
from markupsafe import Markup
from werkzeug.security import safe_join

from incorporealcms.lib import get_meta_str, init_md, instance_resource_path_to_request_path, parse_md, render

logger = logging.getLogger(__name__)

bp = Blueprint('pages', __name__, url_prefix='/')


@bp.route('/', defaults={'path': 'index'})
@bp.route('/<path:path>')
def display_page(path):
    """Get the file contents of the requested path and render the file."""
    try:
        resolved_path, render_type = request_path_to_instance_resource_path(path)
        logger.debug("received request for path '%s', resolved to '%s', type '%s'",
                     path, resolved_path, render_type)
    except PermissionError:
        abort(400)
    except IsADirectoryError:
        return redirect(f'/{path}/', code=301)
    except FileNotFoundError:
        abort(404)

    if render_type == 'file':
        return send_from_directory(app.instance_path, resolved_path)
    elif render_type == 'symlink':
        logger.debug("attempting to redirect path '%s' to reverse of resource '%s'", path, resolved_path)
        redirect_path = f'/{instance_resource_path_to_request_path(resolved_path)}'
        logger.debug("redirect path: '%s'", redirect_path)
        return redirect(redirect_path, code=301)
    elif render_type == 'markdown':
        logger.debug("treating path '%s' as markdown '%s'", path, resolved_path)
        return handle_markdown_file_path(resolved_path)
    else:
        logger.exception("unsupported render_type '%s'!?", render_type)
        abort(500)


def handle_markdown_file_path(resolved_path):
    """Given a location on disk, attempt to open it and render the markdown within."""
    try:
        content, md, page_name, page_title, mtime = parse_md(resolved_path)
    except OSError:
        logger.exception("resolved path '%s' could not be opened!", resolved_path)
        abort(500)
    except ValueError:
        logger.exception("error parsing/rendering markdown!")
        abort(500)
    except TypeError:
        logger.exception("error loading/rendering markdown!")
        abort(500)
    else:
        parent_navs = generate_parent_navs(resolved_path)
        extra_footer = get_meta_str(md, 'footer') if md.Meta.get('footer') else None
        template = get_meta_str(md, 'template') if md.Meta.get('template') else 'base.html'

        # check if this has a HTTP redirect
        redirect_url = get_meta_str(md, 'redirect') if md.Meta.get('redirect') else None
        if redirect_url:
            logger.debug("redirecting via meta tag to '%s'", redirect_url)
            return redirect(redirect_url, code=301)

        return render(template, title=page_title, description=get_meta_str(md, 'description'),
                      image=get_meta_str(md, 'image'), base_url=request.base_url, content=content,
                      navs=parent_navs, mtime=mtime.strftime('%Y-%m-%d %H:%M:%S %Z'),
                      extra_footer=extra_footer)


def request_path_to_instance_resource_path(path):
    """Turn a request URL path to the full page path.

    flask.Flask.open_instance_resource will open a file like /etc/hosts if you tell it to,
    which sucks, so we do a lot of work here to make sure we have a valid request to
    something inside the pages dir.
    """
    # check if the path is allowed
    base_dir = os.path.realpath(f'{app.instance_path}/pages/')
    safe_path = safe_join(base_dir, path)
    # bail if the requested real path isn't inside the base directory
    if not safe_path:
        logger.warning("client tried to request a path '%s' outside of the base_dir!", path)
        raise PermissionError

    verbatim_path = os.path.abspath(safe_path)
    resolved_path = os.path.realpath(verbatim_path)
    logger.debug("base_dir '%s', constructed resolved_path '%s' for path '%s'", base_dir, resolved_path, path)

    # see if we have a real file or if we should infer markdown rendering
    if os.path.exists(resolved_path):
        # if this is a file-like request but actually a directory, redirect the user
        if os.path.isdir(resolved_path) and not path.endswith('/'):
            logger.info("client requested a path '%s' that is actually a directory", path)
            raise IsADirectoryError

        # if the requested path contains a symlink, redirect the user
        if verbatim_path != resolved_path:
            logger.info("client requested a path '%s' that is actually a symlink to file '%s'", path, resolved_path)
            return resolved_path.replace(f'{app.instance_path}{os.path.sep}', ''), 'symlink'

        # derive the proper markdown or actual file depending on if this is a dir or file
        if os.path.isdir(resolved_path):
            resolved_path = os.path.join(resolved_path, 'index.md')
            return resolved_path.replace(f'{app.instance_path}{os.path.sep}', ''), 'markdown'

        logger.info("final DIRECT path = '%s' for request '%s'", resolved_path, path)
        return resolved_path.replace(f'{app.instance_path}{os.path.sep}', ''), 'file'

    # if we're here, this isn't direct file access, so try markdown inference
    verbatim_path = f'{safe_path}.md'
    resolved_path = os.path.realpath(verbatim_path)

    # does the final file actually exist?
    if not os.path.exists(resolved_path):
        logger.warning("requested final path '%s' does not exist!", resolved_path)
        raise FileNotFoundError

    # check for symlinks
    if verbatim_path != resolved_path:
        logger.info("client requested a path '%s' that is actually a symlink to file '%s'", path, resolved_path)
        return resolved_path.replace(f'{app.instance_path}{os.path.sep}', ''), 'symlink'

    logger.info("final path = '%s' for request '%s'", resolved_path, path)
    # we checked that the file exists via absolute path, but now we need to give the path relative to instance dir
    return resolved_path.replace(f'{app.instance_path}{os.path.sep}', ''), 'markdown'


def generate_parent_navs(path):
    """Create a series of paths/links to navigate up from the given resource path."""
    if path == 'pages/index.md':
        # bail and return the domain name as a terminal case
        return [(app.config['DOMAIN_NAME'], '/')]
    else:
        if path.endswith('index.md'):
            # index case: one dirname for foo/bar/index.md -> foo/bar, one for foo/bar -> foo
            parent_resource_dir = os.path.dirname(os.path.dirname(path))
        else:
            # usual case: foo/buh.md -> foo
            parent_resource_dir = os.path.dirname(path)

        # generate the request path (i.e. what the link will be) for this path, and
        # also the resource path of this parent (which is always a dir, so always index.md)
        request_path = f'/{instance_resource_path_to_request_path(path)}'
        parent_resource_path = os.path.join(parent_resource_dir, 'index.md')

        logger.debug("resource path: '%s'; request path: '%s'; parent resource path: '%s'", path,
                     request_path, parent_resource_path)

        # for issues regarding parser reuse (see lib.init_md) we reinitialize the parser here
        md = init_md()

        # read the resource
        try:
            with app.open_instance_resource(path, 'r') as entry_file:
                entry = entry_file.read()
            _ = Markup(md.convert(entry))
            page_name = (" ".join(md.Meta.get('title')) if md.Meta.get('title')
                         else request_path_to_breadcrumb_display(request_path))
            return generate_parent_navs(parent_resource_path) + [(page_name, request_path)]
        except FileNotFoundError:
            return generate_parent_navs(parent_resource_path) + [(request_path, request_path)]


def request_path_to_breadcrumb_display(path):
    """Given a request path, e.g. "/foo/bar/baz/", turn it into breadcrumby text "baz"."""
    undired = path.rstrip('/')
    leaf = undired[undired.rfind('/'):]
    return leaf.strip('/')



"""Generate Atom and RSS feeds based on content in a blog-ish location.

This parses a special root directory, feed/, for YYYYMMDD-foo.md files,
and combines them into an Atom or RSS feed. These files *should* be symlinks
to the real pages, which may mirror the same YYYYMMDD-foo.md file naming scheme
under pages/ (which may make sense for a blog) if they want, but could just
as well be pages/foo content.

SPDX-FileCopyrightText: © 2023 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import logging
import os
import re

from feedgen.feed import FeedGenerator
from flask import Blueprint, Response, abort
from flask import current_app as app

from incorporealcms.lib import instance_resource_path_to_request_path, parse_md

logger = logging.getLogger(__name__)

bp = Blueprint('feed', __name__, url_prefix='/feed')


@bp.route('/<feed_type>')
def serve_feed(feed_type):
    """Serve the Atom or RSS feed as requested."""
    logger.warning("wat")
    if feed_type not in ('atom', 'rss'):
        abort(404)

    fg = FeedGenerator()
    fg.id(f'https://{app.config["DOMAIN_NAME"]}/')
    fg.title(f'{app.config["TITLE_SUFFIX"]}')
    fg.author(app.config["AUTHOR"])
    fg.link(href=f'https://{app.config["DOMAIN_NAME"]}/feed/{feed_type}', rel='self')
    fg.link(href=f'https://{app.config["DOMAIN_NAME"]}', rel='alternate')
    fg.subtitle(f"Blog posts and other dated materials from {app.config['TITLE_SUFFIX']}")

    # get recent feeds
    feed_path = os.path.join(app.instance_path, 'feed')
    feed_entry_paths = [os.path.join(dirpath, filename) for dirpath, _, filenames in os.walk(feed_path)
                        for filename in filenames if os.path.islink(os.path.join(dirpath, filename))]
    for feed_entry_path in sorted(feed_entry_paths):
        # get the actual file to parse it
        resolved_path = os.path.realpath(feed_entry_path).replace(f'{app.instance_path}/', '')
        try:
            content, md, page_name, page_title, mtime = parse_md(resolved_path)
            link = f'https://{app.config["DOMAIN_NAME"]}/{instance_resource_path_to_request_path(resolved_path)}'
        except (OSError, ValueError, TypeError):
            logger.exception("error loading/rendering markdown!")
            abort(500)

        fe = fg.add_entry()
        fe.id(_generate_feed_id(feed_entry_path))
        fe.title(page_name if page_name else page_title)
        fe.author(app.config["AUTHOR"])
        fe.link(href=link)
        fe.content(content, type='html')

    if feed_type == 'atom':
        return Response(fg.atom_str(pretty=True), mimetype='application/atom+xml')
    else:
        return Response(fg.rss_str(pretty=True), mimetype='application/rss+xml')


def _generate_feed_id(feed_entry_path):
    """For a relative file path, generate the Atom/RSS feed ID for it."""
    date = re.sub(r'.*(\d{4})(\d{2})(\d{2}).*', r'\1-\2-\3', feed_entry_path)
    cleaned = feed_entry_path.replace('#', '/').replace('feed/', '', 1).replace(app.instance_path, '')
    return f'tag:{app.config["DOMAIN_NAME"]},{date}:{cleaned}'



"""Test basic configuration stuff.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import os

from incorporealcms import create_app

HERE = os.path.dirname(os.path.abspath(__file__))


def test_config():
    """Test create_app without passing test config."""
    instance_path = os.path.join(HERE, 'instance')
    assert not create_app(instance_path=instance_path).testing
    assert create_app(instance_path=instance_path, test_config={"TESTING": True}).testing


def test_markdown_meta_extension_always():
    """Test that the markdown meta extension is always loaded, even if not specified."""
    app = create_app(instance_path=os.path.join(HERE, 'instance'),
                     test_config={'MARKDOWN_EXTENSIONS': []})
    client = app.test_client()
    response = client.get('/')
    assert response.status_code == 200
    assert b'<title>Index - example.com</title>' in response.data


def test_custom_markdown_extensions_work():
    """Test we can change extensions via config, and that they work.

    This used to test smarty, but that's added by default now, so we test
    that it can be removed by overriding the option.
    """
    app = create_app(instance_path=os.path.join(HERE, 'instance'))
    client = app.test_client()
    response = client.get('/mdash-or-triple-dash')
    assert response.status_code == 200
    assert b'word &mdash; word' in response.data

    app = create_app(instance_path=os.path.join(HERE, 'instance'),
                     test_config={'MARKDOWN_EXTENSIONS': []})
    client = app.test_client()
    response = client.get('/mdash-or-triple-dash')
    assert response.status_code == 200
    assert b'word --- word' in response.data


def test_title_override():
    """Test that a configuration with a specific title overrides the default."""
    instance_path = os.path.join(HERE, 'instance')
    app = create_app(instance_path=instance_path, test_config={'TITLE_SUFFIX': 'suou.net'})
    client = app.test_client()
    response = client.get('/no-title')
    assert response.status_code == 200
    assert b'<title>/no-title - suou.net</title>' in response.data


def test_media_file_access(client):
    """Test that media files are served, and properly."""
    response = client.get('/media/favicon.png')
    assert response.status_code == 200
    assert response.headers['content-type'] == 'image/png'


def test_favicon_override():
    """Test that a configuration with a specific favicon overrides the default."""
    instance_path = os.path.join(HERE, 'instance')
    app = create_app(instance_path=instance_path, test_config={'FAVICON': '/media/foo.png'})
    client = app.test_client()
    response = client.get('/no-title')
    assert response.status_code == 200
    assert b'<link rel="icon" href="/media/foo.png">' in response.data


def test_misconfigured_markdown_extensions():
    """Test that a misconfigured markdown extensions leads to a 500 at render time."""
    instance_path = os.path.join(HERE, 'instance')
    app = create_app(instance_path=instance_path, test_config={'MARKDOWN_EXTENSIONS': 'WRONG'})
    client = app.test_client()
    response = client.get('/no-title')
    assert response.status_code == 500



"""Configure the test application.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""

LOGGING = {
    'version': 1,
    'formatters': {
        'default': {
            'format': '[%(asctime)s %(levelname)-7s %(name)s] %(message)s',
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'default',
        },
    },
    'loggers': {
        'incorporealcms.mdx': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
        'incorporealcms.pages': {
            'level': 'DEBUG',
            'handlers': ['console'],
        },
    },
}



"""Create generic figures with captions.

SPDX-FileCopyrightText: © 2022 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import re
from xml.etree.ElementTree import SubElement  # nosec B405 - not parsing untrusted XML here

import markdown


class FigureExtension(markdown.Extension):
    """Wrap the markdown prepcoressor."""

    def extendMarkdown(self, md):
        """Add FigureBlockProcessor to the Markdown instance."""
        md.parser.blockprocessors.register(FigureBlockProcessor(md.parser), 'figure', 100)


class FigureBlockProcessor(markdown.blockprocessors.BlockProcessor):
    """Process figures."""

    # |> thing to put in the figure
    # |: optional caption for the figure
    # optional whatever else, like maybe an attr_list
    figure_regex = re.compile(r'^[ ]{0,3}\|>[ ]{0,3}(?P<content>[^\n]*)')
    caption_regex = re.compile(r'^[ ]{0,3}\|:[ ]{0,3}(?P<caption>[^\n]*)')

    def test(self, parent, block):
        """Determine if we should process this block."""
        lines = block.split('\n')
        return bool(self.figure_regex.search(lines[0]))

    def run(self, parent, blocks):
        """Replace the top block with HTML."""
        block = blocks.pop(0)
        lines = block.split('\n')

        # consume line and create a figure
        figure_match = self.figure_regex.search(lines[0])
        lines.pop(0)
        content = figure_match.group('content')
        figure = SubElement(parent, 'figure')
        figure.text = content
        if lines:
            if caption_match := self.caption_regex.search(lines[0]):
                # consume line and add the caption as a child of the figure
                lines.pop(0)
                caption = caption_match.group('caption')
                figcaption = SubElement(figure, 'figcaption')
                figcaption.text = caption
        if lines:
            # other lines are mysteries, might be attr_list, so re-append
            # make sure there's a child to hang the rest (which is maybe an attr_list?) off of
            # this is probably a bad hack
            if not len(list(figure)):
                SubElement(figure, 'span')
            rest = '\n'.join(lines)
            figure[-1].tail = f'\n{rest}'


def makeExtension(*args, **kwargs):
    """Provide the extension to the markdown extension loader."""
    return FigureExtension(*args, **kwargs)



"""Error page views for 400, 404, etc.

SPDX-FileCopyrightText: © 2021 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
from incorporealcms.lib import render


def bad_request(error):
    """Display 400 error messaging."""
    return render('400.html'), 400


def internal_server_error(error):
    """Display 500 error messaging."""
    return render('500.html'), 500


def page_not_found(error):
    """Display 404 error messaging."""
    return render('404.html'), 404



"""Default configuration.

SPDX-FileCopyrightText: © 2020 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""


class Config(object):
    """Represent the default configuration.

    Reminder: this should be overwritten in the instance config.py, not here!
    """

    DEBUG = False
    TESTING = False

    LOGGING = {
        'version': 1,
        'formatters': {
            'default': {
                'format': '[%(asctime)s %(levelname)-7s %(name)s] %(message)s',
            },
        },
        'handlers': {
            'console': {
                'level': 'DEBUG',
                'class': 'logging.StreamHandler',
                'formatter': 'default',
            },
        },
        'loggers': {
            '': {
                'level': 'INFO',
                'handlers': ['console'],
            },
        },
    }

    MARKDOWN_EXTENSIONS = ['extra', 'incorporealcms.mdx.figures', 'sane_lists', 'smarty', 'toc']
    MARKDOWN_EXTENSION_CONFIGS = {
        'extra': {
            'footnotes': {
                'UNIQUE_IDS': True,
            },
        },
        'smarty': {
            'smart_dashes': True,
            'smart_quotes': False,
            'smart_angled_quotes': False,
            'smart_ellipses': True,
        },
    }

    MEDIA_DIR = 'media'

    # customizations
    PAGE_STYLES = {
        'dark': '/static/css/dark.css',
        'light': '/static/css/light.css',
        'plain': '/static/css/plain.css',
    }

    DEFAULT_PAGE_STYLE = 'light'
    DOMAIN_NAME = 'example.com'
    TITLE_SUFFIX = DOMAIN_NAME
    CONTACT_EMAIL = 'admin@example.com'

    # feed settings
    AUTHOR = {'name': 'Test Name', 'email': 'admin@example.com'}

    # specify FAVICON in your instance config.py to override the provided icon



"""Test graphviz functionality.

SPDX-FileCopyrightText: © 2021 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import os

from incorporealcms import create_app

HERE = os.path.dirname(os.path.abspath(__file__))


def app_with_pydot():
    """Create the test app, including the pydot extension."""
    return create_app(instance_path=os.path.join(HERE, 'instance'),
                      test_config={'MARKDOWN_EXTENSIONS': ['incorporealcms.mdx.pydot']})


def test_functional_initialization():
    """Test initialization with the graphviz config."""
    app = app_with_pydot()
    assert app is not None


def test_graphviz_is_rendered():
    """Initialize the app with the graphviz extension and ensure it does something."""
    app = app_with_pydot()
    client = app.test_client()

    response = client.get('/test-graphviz')
    assert response.status_code == 200
    assert b'~~~pydot' not in response.data
    assert b'data:image/png;base64' in response.data


def test_two_graphviz_are_rendered():
    """Test two images are rendered."""
    app = app_with_pydot()
    client = app.test_client()

    response = client.get('/test-two-graphviz')
    assert response.status_code == 200
    assert b'~~~pydot' not in response.data
    assert b'data:image/png;base64' in response.data


def test_invalid_graphviz_is_not_rendered():
    """Check that invalid graphviz doesn't blow things up."""
    app = app_with_pydot()
    client = app.test_client()

    response = client.get('/test-invalid-graphviz')
    assert response.status_code == 500
    assert b'INTERNAL SERVER ERROR' in response.data


def test_figures_are_rendered(client):
    """Test that a page with my figure syntax renders as expected."""
    response = client.get('/figures')
    assert response.status_code == 200
    assert (b'<figure class="right"><img alt="fancy captioned logo" src="bss-square-no-bg.png" />'
            b'<figcaption>this is my cool logo!</figcaption></figure>') in response.data
    assert (b'<figure><img alt="vanilla captioned logo" src="bss-square-no-bg.png" />'
            b'<figcaption>this is my cool logo without an attr!</figcaption>\n</figure>') in response.data
    assert (b'<figure class="left"><img alt="fancy logo" src="bss-square-no-bg.png" />'
            b'<span></span></figure>') in response.data
    assert b'<figure><img alt="just a logo" src="bss-square-no-bg.png" /></figure>' in response.data



"""Serve static files from the instance directory.

SPDX-FileCopyrightText: © 2022 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
import os

from flask import Blueprint
from flask import current_app as app
from flask import send_from_directory

bp = Blueprint('static', __name__, url_prefix='/custom-static')


@bp.route('/<path:name>')
def serve_instance_static_file(name):
    """Serve a static file from the instance directory, used for customization."""
    return send_from_directory(os.path.join(app.instance_path, 'custom-static'), name)



"""Test the feed methods.

SPDX-FileCopyrightText: © 2023 Brian S. Stephan <bss@incorporeal.org>
SPDX-License-Identifier: AGPL-3.0-or-later
"""
from incorporealcms.feed import serve_feed


def test_unknown_type_is_404(client):
    """Test that requesting a feed type that doesn't exist is a 404."""
    response = client.get('/feed/wat')
    assert response.status_code == 404


def test_atom_type_is_200(client):
    """Test that requesting an ATOM feed is found."""
    response = client.get('/feed/atom')
    assert response.status_code == 200
    assert 'application/atom+xml' in response.content_type
    print(response.text)


def test_rss_type_is_200(client):
    """Test that requesting an RSS feed is found."""
    response = client.get('/feed/rss')
    assert response.status_code == 200
    assert 'application/rss+xml' in response.content_type
    print(response.text)


def test_feed_generator_atom(app):
    """Test the root feed generator."""
    with app.test_request_context():
        content = serve_feed('atom')
    assert b'<id>https://example.com/</id>' in content.data
    assert b'<email>admin@example.com</email>' in content.data
    assert b'<name>Test Name</name>' in content.data


def test_feed_generator_rss(app):
    """Test the root feed generator."""
    with app.test_request_context():
        content = serve_feed('rss')
    assert b'<author>admin@example.com (Test Name)</author>' in content.data
